#Generic communication functions

import struct

from array import array
from datetime import date
from fcntl import ioctl
from os import kill, listdir, stat, remove
from os.path import exists, getctime
from shutil import move
from signal import SIGTERM
from socket import socket, AF_INET, SOCK_DGRAM
from subprocess import Popen
from time import sleep
from types import TupleType

import gwutil
from variables import sshServers, webServers, SPECIAL_FILE, POWER_UPLOAD, \
    POWER_DOWNLOAD, BASE_DIR, QUEUED_DIR, SMALL_FILE_SIZE, ARCHIVE_DIR, \
    COMMS_CONNECT_TIME, COMMS_STAY_AWAKE_FILE, OVERIDE_FILE, DATA_DIR, \
    GEOPHONE_SPECIAL_FILE, PROBE_SPECIAL_FILE
from gwschedule import get_state_from_file
from gwlogging import logger, gw_stdout, gw_stderr
from gwcommands import IFDOWN, IFUP, PING, SCP, SSH, WGET


def chooseServer(servers, gprs=False):
    """Cycles through the dictionary of servers and returns the name
    of first one which can be sucessfully pinged otherwise returns none"""
    logger.debug("Chooose server")
    logger.info("GPRS connection %s" % gprs)
    for name, server in servers.items():
        #check if server is only accesible via any connection, if so try ping
        # or if only available via gprs check using gprs connection
        if (not server.gprs_only
                or (gprs and server.gprs_only)):
            cmd = "%s -c 3 %s" % (PING, server.address)
            logger.debug(cmd)
            p = Popen(args=cmd,
                shell=True, stdout=gw_stdout, stderr=gw_stderr)
            rc = p.wait()
            if rc == 0:
                logger.debug("Server chosen: %s" % name)
                return name
    return None


def chooseWebServer(sshServer=None):
    """Checks to see if the chosen ssh server is also a valid web server,
    if it is then use that, if not find a valid server"""
    if sshServer is not None and sshServer in webServers.keys():
        logger.debug("Using same server as SSH for web")
        return sshServer
    return chooseServer(webServers)


def chooseSshServer(webServer=None, gprs=False):
    """Checks to see if the chosen web server is also a valid ssh server,
    if it is then use that, if not find a valid server"""
    if (webServer is not None
            and webServer in sshServers.keys()
            and (gprs or not sshServers[webServer].gprs_only)):
        logger.debug("using web server for ssh")
        return webServer
    return chooseServer(sshServers, gprs)

def getProbeSpecial(server_name):
    local_special = "%s%s/%s" % (BASE_DIR, DATA_DIR, PROBE_SPECIAL_FILE)
    getSpecial(server_name, PROBE_SPECIAL_FILE, local_special)

def getGeophoneSpecial(server_name):
    local_special = "%s%s/%s" % (BASE_DIR, DATA_DIR, GEOPHONE_SPECIAL_FILE)
    getSpecial(server_name, GEOPHONE_SPECIAL_FILE, local_special)

def getMainSpecial(server_name, mode):
    remote_special = "/%s_%s" % (mode, SPECIAL_FILE)
    local_special = "%s%s/%s" % (BASE_DIR, DATA_DIR, SPECIAL_FILE)
    getSpecial(server_name, remote_special, local_special)

def getSpecial(server_name, remote_fname, local_fname):
    """Gets the file from the specified server"""
    logger.debug("Remote file = %s" % remote_fname)
    server = webServers[server_name]
    url = server.address + server.webPath + "/" + remote_fname
    logger.info("Fetching %s" % url)
    wget = Popen(args="%s -q  -O %s %s" % 
        (WGET, local_fname, url),
        shell=True, stdout=gw_stdout, stderr=gw_stderr)
    wget.wait()
    if stat(local_fname).st_size == 0:
        logger.warn("Special file is empty; removing")
        remove(local_fname)


def sendPowerState(server_name, mode, state):
    server = webServers[server_name]
    # Send our power state home
    logger.info("Sending power state home")
    wget = Popen(
        args="%s -q -O /dev/null -q %s%s%s?id=%s\&state=%i" % (WGET,
        server.address, server.webPath, POWER_UPLOAD, mode, state),
        shell=True, stdout=gw_stdout, stderr=gw_stderr)
    wget.wait()


def fetchOverridePowerState(server_name, mode):
    server = webServers[server_name]
    # Get override power state
    logger.info("Fetching overide power state")
    wget = Popen(args="%s -O %s -q %s%s%s?id=%s" % (WGET, 
        OVERIDE_FILE, server.address, server.webPath, POWER_DOWNLOAD, mode),
                shell=True, stdout=gw_stdout,stderr=gw_stderr)
    wget.wait()


def uploadFiles(server_name, force_all=True, connection=None):
    """ Upload the files that are queued to the server specified"""
    server = sshServers[server_name]
    # Copy  the tars across:
    logger.info("Sending queued files to %s:%s" % (server.address,
        server.uploadPath))
    for fname in listdir(BASE_DIR + QUEUED_DIR):
        size = gwutil.file_size(BASE_DIR + QUEUED_DIR + "/" + fname)
        if (force_all) or (connection == "WIFI") or (size < SMALL_FILE_SIZE):
            # Copy the data:
            sendFile(server_name, fname)
        elif(date.fromtimestamp(getctime(fname)) == date.today()):
            logger.info("Sending todays file regards of size")
            sendFile(server_name, fname)
        else:
            logger.warn("%s too big at %s" % (fname, str(size)))


def sendFile(server_name, fname):
    server = sshServers[server_name]
    cmd = "%s -B -i %s %s %s@%s:%s" % (SCP, server.keyPath,
            fname, server.user, server.address, server.uploadPath)
    logger.info(cmd)
    p = Popen(args=cmd, cwd=(BASE_DIR + QUEUED_DIR), shell=True,
                stdout=gw_stdout, stderr=gw_stderr)
    r = p.wait()

    # Archive the file if the scp was successful
    if r == 0:
        logger.info("Sent %s" % fname)
        move(BASE_DIR + QUEUED_DIR + "/" + fname, BASE_DIR + ARCHIVE_DIR)            
    else:
        logger.error("Sending %s failed" % fname)
    

def open_ssh_tunnel(server_name, reverse_port):
    server = sshServers[server_name]
    # Determine comms window length based on batv state
    comms_window_time = COMMS_CONNECT_TIME * get_state_from_file()
    # Establish reverse tunnel.
    logger.info("Establishing reverse tunnel to %s:%i" % (server.address,
        reverse_port))
    cmd = "%s -i %s %s@%s -R %i:localhost:22 sleep %s" % (SSH,
        server.keyPath, server.user, server.address, reverse_port,
        comms_window_time)
    ssh = Popen(args=cmd,
                shell=True,
                stdout=open("/dev/null", "w"),
                stderr=gw_stderr)

    logger.info("Staying awake for %i seconds" % comms_window_time)
    sleep(comms_window_time)

    # Stay connected whilst comms stay awake file exists.
    while exists(COMMS_STAY_AWAKE_FILE):
        logger.info("Staying connected because of comms stay awake file")
        sleep(10)

    # Comms window closed, remove reverse tunnel.
    kill(ssh.pid, SIGTERM)
    ssh.wait()

#Interface handling code


def interfaces():
    """Gets a list of all network interfaces that are up
    Grabbed from:
    http://aspn.activestate.com/ASPN/Cookbook/Python/Recipe/439093"""
    max_possible = 128  # arbitrary. raise if needed.
    max_bytes = max_possible * 32
    s = socket(AF_INET, SOCK_DGRAM)
    names = array('B', '\0' * max_bytes)
    outbytes = struct.unpack('iL', ioctl(
        s.fileno(),
        0x8912,  # SIOCGIFCONF
        struct.pack('iL', max_bytes, names.buffer_info()[0])))[0]
    namestr = names.tostring()
    return [namestr[i:i + 32].split(
               '\0', 1)[0] for i in range(0, outbytes, 32)]


def is_interface_up(ifname):
    """checks if an interface exists and if it does if it is up"""
    if not net_exists(ifname):
        logger.error("Interface doesn't exist")
        return False

    ## {{{ http://code.activestate.com/recipes/65449/ (r1)
    # set some symbolic constants
    SIOCGIFFLAGS = 0x8913
    null256 = '\0' * 256

    # create a socket so we have a handle to query
    s = socket(AF_INET, SOCK_DGRAM)

    # call ioctl() to get the flags for the given interface
    result = ioctl(s.fileno(), SIOCGIFFLAGS, ifname + null256)

    # extract the interface's flags from the return value
    flags, = struct.unpack('H', result[16:18])

    # check "UP" bit and print a message
    up = (flags & 1) == 1
    return up


def interface_up(ifname):
    """ starts an interface"""
    if not net_exists(ifname):
        logger.error("Interface doesn't exist so cannot up it")
        return False
        #Network exists but isn't up

    if is_interface_up(ifname):
        logger.info("Interface already up")
        interface_down(ifname)
        logger.debug("Interface taken down")
    logger.debug("ifup interface %s" % ifname)
    up = Popen(args="%s %s" % (IFUP, ifname),
        shell=True,
        stdout=gw_stdout,
        stderr=gw_stderr)
    if up.wait() == 0:
        logger.debug("executed sucessfully")
    return is_interface_up(ifname)


def interface_down(ifname):
    if not net_exists(ifname):
        logger.error("Interface doesn't exist")
        return False
    else:
        logger.debug("ifdown interface %s" % ifname)
        down = Popen(args="%s %s" %(IFDOWN, ifname),
            shell=True,
            stdout=gw_stdout,
            stderr=gw_stderr)
        if down.wait() == 0:
            logger.debug("executed sucessfully")
        return not is_interface_up(ifname)


def net_exists(ifname):
    """Returns True if the given interface name exists"""
    return exists("/sys/class/net/%s" % ifname)


def conn_info(pid_file):
    """Returns a tuple containing:
       0) The PID of pppd
       1) The interface name.
       If there's no pppd instance, returns False"""
    if not exists(pid_file):
        return False

    try:
        f = open(pid_file)
        pid = int(f.readline().strip())
        ifname = f.readline().strip()
        return(pid, ifname)
    except:
        return False
    finally:
        f.close()


def disconnect_ppp(pid_file):
    """Disconnects and waits for pppd to terminate"""
    ci = conn_info(pid_file)

    if type(ci) != TupleType:
        # pppd isn't running
        return

    pid, ifname = ci

    kill(pid, SIGTERM)

    while exists(pid_file):
        sleep(1)


def ppp_is_up(pid_file):
    """Returns True when the connection is up"""
    ci = conn_info(pid_file)

    if type(ci) != TupleType:
        return False

    pid, ifname = ci

    if ifname in interfaces():
        return True
    return False
