from datetime import datetime
from serial import Serial

from variables import MSP_SERIAL_PORT, MSP_BAUD_RATE, MSP_RETRY_ATTEMPTS
from gwlogging import logger 
from iceCapeMapping import *
import MspJob
import MspCmd
import MspData
from gwtimeutils import totimestamp, timedelta2seconds, next_time

class iceCape:
    """A class to handle all communication with the iceCape board,
    as designed by Philip Basford 2012, based on gumInter by Graeme Bragg
    2011 """
    version = 2.3
    serial_port = None
    port_name = None
    baud = None

    def __init__(self, port=MSP_SERIAL_PORT, baud=MSP_BAUD_RATE):
        self.port_name = port
        self.baud = baud
        return

    def __del__(self):
        return

    def open(self):
        """Open the serial port for communication"""
        self.serial_port = Serial(self.port_name,
            self.baud, timeout=0.5)

    def close(self):
        """Close the serial port"""
        self.serial_port.close()

    def _send_command(self, cmd):
        self.open()
        cmd_string = "%s\t%s\t%s\t%s\t%s\t%s\t%s\n" % (cmd.cmd,
            cmd.cmdType,
            totimestamp(cmd.exec_time),
            timedelta2seconds(cmd.job_interval),
            timedelta2seconds(cmd.power_time),
            cmd.power_mask,
            cmd.sensor_mask)
        logger.msp(cmd_string)
        self.serial_port.write(cmd_string)
        resp = self.serial_port.readlines(None)
        logger.msp(resp)
        attempts = 1
        while (not ('k\r\n' in resp)
                and (attempts <= MSP_RETRY_ATTEMPTS)):
            logger.warn("Attempt %i failed" % attempts)
            logger.debug("Sending new lines")
            self.serial_port.write("\n\n")
            self.serial_port.readlines(None)
            logger.debug("Retrying command")
            self.serial_port.write(cmd_string)
            resp = self.serial_port.readlines(None)
            attempts += 1
        logger.debug("Command sent")
        self.close()
        clean = [l for l in resp if not l.startswith('#')]
        if "k\r\n" in clean:
            clean.remove("k\r\n")
        if "\r" in clean:
            clean.remove("\r")
        return clean

    #Get version from the MSP
    def get_version(self):
        cmd = MspCmd.MspCmd(MspCmd.C_READ_VERSION)
        resp = self._send_command(cmd)
        s = str.split(resp[0], ":")
        return s[1]

    def get_local_version(self):
        """Get the version of this code"""
        return self.version

    def set_datetime(self, time=datetime.utcnow()):
        cmd = MspCmd.MspCmd(MspCmd.C_EPOCH_SET, totimestamp(time))
        self._send_command(cmd)

    def set_epoch(self, time=datetime.utcnow()):
        self.set_datetime(time)

    def get_datetime(self):
        cmd = MspCmd.MspCmd(MspCmd.C_EPOCH_READ)
        resp = self._send_command(cmd)
        return datetime.utcfromtimestamp(float(resp[0]))

    def get_epoch(self):
        return self.get_datetime()

    #Power Commands

    def power_off_gumstix(self, delay=0):
        self.power_off_beagle(delay)

    def power_off_beagle(self, delay=0):
        """Turn off the gumstix, if no parameter passsed
        then turn off immediately"""
        cmd = MspCmd.MspCmd(MspCmd.C_POWER_DOWN_GUM, delay)
        self._send_command(cmd)

    def power_down(self, delay=0):
        """Kills the Gum AND all peripherals"""
        cmd = MspCmd.MspCmd(MspCmd.C_POWER_DOWN, delay)
        self._send_command(cmd)

    def turn_off_power_output(self, mask):
        self.output_off(mask)

    def turn_on_power_output(self, mask):
        self.output_on(mask)

    def turn_off_sensor_output(self, mask):
        self.output_off(mask)

    def turn_on_sensor_output(self, mask):
        self.output_on(mask)

    def output_on(self, mask):
        cmd = MspCmd.MspCmd(MspCmd.C_OUTPUT_ON, mask)
        self._send_command(cmd)

    def output_off(self, mask):
        cmd = MspCmd.MspCmd(MspCmd.C_OUTPUT_OFF, mask)
        self._send_command(cmd)

    def debug_output_on(self, mask):
        cmd = MspCmd.MspCmd(MspCmd.C_LED_ON, mask)
        self._send_command(cmd)

    def debug_output_off(self, mask):
        cmd = MspCmd.MspCmd(MspCmd.C_LED_OFF, mask)
        self._send_command(cmd)

    def output_status(self):
        cmd = MspCmd.MspCmd(MspCmd.C_OUTPUT_STATUS)
        resp = self._send_command(cmd)
        return int(resp[0])

    def debug_status(self):
        cmd = MspCmd.MspCmd(MspCmd.C_DEBUG_STATUS)
        resp = self._send_command(cmd)
        return int(resp[0])

    def bone_status(self):
        cmd = MspCmd.MspCmd(MspCmd.C_BONE_STATUS)
        resp = self._send_command(cmd)
        return int(resp[0]) == 1

    #Sensor commands
    def read_sensors(self, mask):  # needs checking
        if mask is not None:
            cmd = MspCmd.MspCmd(MspCmd.C_SENSOR_READ, mask)
            return int(self._send_command(cmd)[0])
        else:
            logger.error("Must set a mask")

    #Job commands
    def clear_job_table(self):
        """Clear the MSP job table"""
        cmd = MspCmd.MspCmd(MspCmd.C_JOB_TB_CLEAR)
        self._send_command(cmd)

    def read_job_table(self):
        """Read the current list of jobs from the MSP"""
        cmd = MspCmd.MspCmd(MspCmd.C_JOB_TB_READ)
        resp = self._send_command(cmd)
        return MspJob.parse(resp)

    def read_number_of_jobs(self):
        """Read the current list of jobs from the MSP"""
        cmd = MspCmd.MspCmd(MspCmd.C_JOB_NUM_READ)
        resp = self._send_command(cmd)
        return int(resp[0])

    def add_job(self, jobType, exec_time, interval, power_time,
            power_mask, sensor_mask):
        """Add a job to the MSP"""

        n_time = next_time(exec_time, interval)
        cmd = MspCmd.MspCmd(MspCmd.C_JOB_TB_ADD, jobType, n_time,
            interval, power_time, power_mask, sensor_mask)
        self._send_command(cmd)

    def disable_job(self, number):
        """Diasble a job execution"""
        cmd = MspCmd.MspCmd(MspCmd.C_JOB_DISABLE, number)
        self._send_command(cmd)

    def enable_job(self, number):
        """Enable a job for execution"""
        cmd = MspCmd.MspCmd(MspCmd.C_JOB_ENABLE, number)
        self._send_command(cmd)

    #Data commands
    def clear_data_table(self):
        """Clear the MSP data tables"""
        cmd = MspCmd.MspCmd(MspCmd.C_DATA_CLEAR)
        self._send_command(cmd)

    def read_data(self):
        """Read the  data stored on the MSP"""
        cmd = MspCmd.MspCmd(MspCmd.C_DATA_READ)
        resp = self._send_command(cmd)
        return MspData.parse(resp)

    def read_data_amount(self):
        cmd = MspCmd.MspCmd(MspCmd.C_DATA_NUM_READ)
        resp = self._send_command(cmd)
        return int(resp[0])
